from .codex import Codex as NPCodex
from .codex import (
    NUMPY_FASTA_ENCODEX,
    NUMPY_FASTA_ENCODEX_MATRIX,
    NUMPY_FASTA_DECODEX,
    INCLUDE_URACIL,
    INCLUDE_REPEAT,
    ENCODING_a,
    ENCODING_c,
    ENCODING_g,
    ENCODING_t,
    ENCODING_u,
    ENCODING_repeat,
)

from .utils import (
    make_encodex_matrix,
    invert_encodex,
    find_row,
    lookup_channel_index,
    get_channel_indices
)



from .functions import getfasta, merge, complement
from ..terminal.tools import process

function_map = {
    'getfasta': getfasta,
    'merge': merge,
    'complement': complement
}


def bedtools(
    subfunction:str,
    stdin:str,
    encoding:str='utf-8',
    **kwargs
):
    '''
    Arguments:
        stdin (str): the stringified bedfile to be passed to `process`
            called with `subfunction`

        **kwargs (dict): the configurable options for `subfunction`.

    Returns:
        stdout (str): the stringified output bedfile.

    '''
    command = function_map[subfunction](**kwargs)
    encoded = stdin.encode(encoding)
    stdout, stderr = process(command, stdin=encoded)
    decoded = stdout.decode(encoding)
    return decoded



import inspect, sys

def variable_name(variable):
    '''
    Argument:
        variable: the variable to search for from outer-most frame inner-wards

    Returns:
        name (str): the string characters corresponding to the name of the
            variable
    '''
    for fi in reversed(inspect.stack()):
        names = [
            var_name
            for var_name, var_val in fi.frame.f_locals.items()
            if var_val is variable
        ]
        if len(names) > 0:
            return names[0]


def error_message(command, error='', stdout_data=None, stderr_data=None)->None:
    '''
    Arguments:
        command (list): a list of strings indiciating the command and its
            arguments

        error (Error): Optional. By default ''. The error thrown.

        stdout_data: Optional. By default None. If provided, writes to
            sys.stderr.

        stderr_data: Optional. By default None. If provided, writes to
            sys.stderr.

    Returns:
        None
    '''

    msg = '\nERROR during execution of {cmd}:\t{err}\n'.format(cmd=' '.join(command), err=str(error))
    print(msg, file=sys.stderr)
    if stdout_data is not None:
        print("\nStandard output:\n----------------", file=sys.stderr)
        print(stdout_data, file=sys.stderr)
    if stderr_data is not None:
        print("\nStandard output:\n----------------", file=sys.stderr)
        print(stderr_data, file=sys.stderr)
    return



import os
from multiprocessing import Pool
from .defaults import (LABEL_ORDER, USE_OTHER_CLASS, OTHER_CLASS)
# from ntai.ranges import LabeledRange, LabeledRanges
from lrng import LabeledRange, LabeledRanges


from lrng.numba import (
    coalesce as nb_coalesce,
    label_range as nb_label_range,
    relevant_labels as nb_relevant_labels
)
from lrng.numpy import (
    coalesce as np_coalesce,
    label_range as np_label_range,
    relevant_labels as np_relevant_labels
)

def coalesce(which='numpy'):
    return np_coalesce if which == 'numpy' else nb_coalesce
def label_range(which='numpy'):
    return np_label_range if which == 'numpy' else nb_label_range
def relevant_labels(which='numpy'):
    return np_relevant_labels if which == 'numpy' else nb_relevant_labels

class Labeler:
    def __init__(
        self,
        label_order: str = LABEL_ORDER,
        use_other_class: bool = USE_OTHER_CLASS,
        other_class: str = OTHER_CLASS,
        processes: int = 1,
        use_numba: bool=False
    ):
        '''
        Arguments:
            label_order (str): the order labels should be embedded.
            use_other_class (bool): whether or not a class `"Other"` should be
                used.
            other_class (str): the class name of `"Other"`
            processes (int): number of CPU cores to use when encoding / decoding.
                By default 1. If set to None, uses all.
        '''
        if processes is None: processes = os.cpu_count()
        self.processes = processes
        self.use_other_class = use_other_class
        self.other_class = other_class

        if (use_other_class):
            if other_class not in label_order:
                label_order += [other_class]
        self.label_order = label_order
        self.use_numba = use_numba

    def encode(self, sequence:str, ranges, offset:int=0) -> list:
        '''
        Arguments:
            sequence (str): the sequence to embed.
            ranges (LabeledRanges): the reference ranges to use to generate the
                embedding of the sequence.
            offset (int): how far to offset each index. By default `0`.
        Returns:
            embedding (list): the embedded sequence.
        '''
        if isinstance(ranges, LabeledRanges):
            ranges = ranges.as_list()
        lr = label_range('numba' if self.use_numba else 'numpy')
        return lr(offset, offset+len(sequence), ranges, self.label_order, self.use_other_class)

    def label(self, sequence:list, reference_labels:dict):
        '''
        Arguments:
            sequence (list): a list consisting of the at least the bed6
                information of the sequence e.g.
                    `[chromosome, start, stop, name, score, strand,]`

            reference_labels (dict): a dctionary consisting the following
                structure:
                    ```
                    {
                        <chromosome>: {
                            <strand>: LabeledRanges(...),
                            ...
                        }, ...
                    }
                    ```
        Returns:
            sequence_ranges (LabeledRanges): the ranges for which a given class
                appear in the provided sequence
        '''
        chromosome, start, stop, name, score, strand, *_ = sequence
        reference_ranges = reference_labels[chromosome][strand]
        if isinstance(reference_ranges, LabeledRanges):
            reference_ranges = reference_ranges.as_list()
        rl = relevant_labels('numba' if self.use_numba else 'numpy')
        result = rl(int(start), int(stop), reference_ranges, self.label_order)
        return LabeledRanges(result)



def bed_str_to_list(bed:list):
    return [line.split('\t') for line in bed.rstrip('\n').split('\n')]

def bed_list_to_str(bed:str):
    return '\n'.join(['\t'.join([str(el) for el in sub]) for sub in bed])



import os, sys, uuid
from ..terminal.tools import arg

def complement(
    i:str,
    g:str,
    L:bool=False,
    s:bool=False,
):
    '''
    Arguments:
        i (str): input file.
        g (str): the genome file.
        L (bool): Limit the output to solely the chromosomes that are
            represented in the -i file. Chromosomes that are in -g but not -i
            will be suppressed.
        s (bool): Force strandedness. That is, only merge features that are the
            same strand. By default, this is disabled.


    '''
    arg_i = arg('i', i)
    arg_g = arg('g', g)
    arg_L = arg('L', L, flag=True)
    arg_s = arg('s', s, flag=True)

    command = [
        'bedtools',
        'complement',
        *arg_i, *arg_g, *arg_L, *arg_s,
    ]
    return command

def merge(
    i:str,
    s:bool=False,
    S:str=None,
    d:int=0,
    c:str=None,
    o:str=None,
    header:bool=False,
    delim:str=';'
):
    '''
    Arguments:
        i (str): input file.

        s (bool): Force strandedness. That is, only merge features that are the
            same strand. By default, this is disabled.

        S (str): Force merge for one specific strand only. Follow with + or - to
            force merge from only the forward or reverse strand, respectively.
            By default, merging is done without respect to strand.

        d (int): Maximum distance between features allowed for features to be
            merged. Default is 0. That is, overlapping and/or book-ended
            features are merged.

        c (str): specify columns from the input file to operate upon
            (see -o option, below). Multiple columns can be specified in a
            comma-delimited list.

        o (str): Specify the operation that should be applied to -c.
            Valid operations:
                sum, min, max, absmin, absmax, mean, median,
                collapse (i.e., print a delimited list (duplicates allowed)),
                distinct (i.e., print a delimited list (NO duplicates allowed)),
                count
                count_distinct (i.e., a count of the unique values in the column),
            Default: sum
                Multiple operations can be specified in a comma-delimited list.
                If there is only column, but multiple operations, all operations
                will be applied on that column. Likewise, if there is only one
                operation, but multiple columns, that operation will be applied
                to all columns. Otherwise, the number of columns must match the
                number of operations, and will be applied in respective order.

                E.g., -c 5,4,6 -o sum,mean,count will give the sum of column 5,
                the mean of column 4, and the count of column 6.
                The order of output columns will match the ordering given in the
                command.

        header (bool): Print the header from the A file prior to results.

        delim (str): Specify a custom delimiter for the -nms and -scores concat
            options
                Example: -delim "|"
                Default: ";"

    Returns:
        command (list): a list of strings containing the bedtools getfasta
            command
    '''
    arg_i = arg('i', i)
    arg_s = arg('s', s, flag=True)
    arg_S = arg('S', S)
    arg_d = arg('d', d)
    arg_c = arg('c', c)
    arg_o = arg('o', o)
    arg_header = arg('header', header, flag=True)
    arg_delim  = arg('delim', delim)

    command = [
        'bedtools',
        'merge',
        *arg_i, *arg_s, *arg_S,
        *arg_d, *arg_c, *arg_o,
        *arg_header, *arg_delim
    ]
    return command



def getfasta(
    fi:     str,
    bed:    str = '/dev/stdin',
    fo:     str = '/dev/stdout',
    name:   bool = False,
    tab:    bool = False,
    bedOut: bool = False,
    s:      bool = False,
    split:  bool = False
) -> list:
    '''
    Arguments:
        fi (str): input FASTA

        bed (str): <BED/GFF/VCF>

        fo (str): Specify an output file name. By default, output goes to stdout

        name (bool): Use the “name” column in the BED file for the FASTA headers
            in the output FASTA file.

        tab (bool): Report extract sequences in a tab-delimited format instead
            of in FASTA format.

        bedOut (bool): Report extract sequences in a tab-delimited BED format
            instead of in FASTA format.

        s (bool): Force strandedness. If the feature occupies the antisense
            strand, the sequence will be reverse complemented. Default: strand
            information is ignored.

        split (bool): Given BED12 input, extract and concatenate the sequences
            from the BED “blocks” (e.g., exons)
    Returns:
        command (list): a list of strings containing the bedtools getfasta
            command
    '''
    arg_fi = arg('fi', fi)
    arg_bed = arg('bed', bed)
    arg_fo = arg('fo', fo)
    arg_name = arg('name', name, flag=True)
    arg_tab = arg('tab', tab, flag=True)
    arg_bedOut = arg('bedOut', bedOut, flag=True)
    arg_s = arg('s', s, flag=True)
    arg_split = arg('split', split, flag=True)
    command = [
        'bedtools',
        'getfasta',
        *arg_fi, *arg_bed, *arg_fo, *arg_name,
        *arg_tab, *arg_bedOut, *arg_s, *arg_split,
    ]
    return command



import os, sys, subprocess, uuid
from .utils import variable_name, error_message
from .popen import POPEN_DEFAULTS

def arg(name, val, short=True, equal=False, flag=False):
    dashes = '-' if short else '--'
    sep = '=' if equal else ' '

    arg_str = ''

    if val is not None:
        if not flag:
            arg_str = '{}{}{}{}'.format(dashes, name, sep, val)
        else:
            if val:
                arg_str = '{}{}'.format(dashes, name)

    return arg_str.split(' ') if arg_str != '' else []

def clean_command(command:list):
    cleaned = [
        sub
        for part in command
        for sub in part.split(' ')
        if sub != ''
    ]
    return cleaned



def process(command:list, stdin:str, popen_options={}):
    '''
    Arguments:
        command (list): a list of strings indicating the command and its
            arguments to spawn as a subprocess.

        stdin (str): passed as stdin to the subprocess. Assumed to be utf-8
            encoded.

        popen_options (dict): used to configure the subprocess.Popen command

    Returns:
        stdout, stderr
    '''
    command = clean_command(command)
    popen_config = POPEN_DEFAULTS.copy()
    popen_config.update(popen_options)
    try:
        pid = subprocess.Popen(args = command, **popen_config)
        stdout_data, stderr_data = pid.communicate(stdin)
    except OSError as err:
        error_message(command, err)
        sys.exit(1)
    if pid.returncode != 0:
        error_message(command, 'pid code {}'.format(pid.returncode), stdout_data, stderr_data)
    return stdout_data, stderr_data



from random import choices, random, randint
from .defaults import FASTA_CHARS
def random_fasta(min_length:int=0, max_length:int=100, repeat_likelihood:float=0.5) -> str:
    '''
    Arguments:
        min_length (int): minimum length the sequence should be.
        max_length (int): maximum length the sequence should be.
        repeat_likelihood (float): likelihood a given character in the sequence
            should be a repeated masked region.
    Returns:
        sequence (str): the fasta sequence.
    '''
    n = randint(min_length, max_length)
    chars = choices(list(FASTA_CHARS), k=n)
    return ''.join([e if random() < repeat_likelihood else e.upper() for e in chars])



from .codex import Codex, NPCodex
from .bedtools import bedtools
from .fetch import fetch_files
from .labeler import Labeler
# from .ranges import LabeledRange, LabeledRanges

name = 'ntai'

major = 0
minor = 0
patch = 16
version = '{}.{}.{}'.format(major, minor, patch)

description = 'Nucleotide A.I.'



import os
from multiprocessing import Pool
from .defaults import (
    REPEAT_CHAR, FASTA_CHARS,
    ENCODE_ORDER,
    INCLUDE_URACIL, INCLUDE_REPEAT,
    FASTA_ENCODEX, FASTA_DECODEX
)

from .utils import (is_fasta, is_uracil)

class Codex:

    def __init__(
        self,
        include_uracil:bool = INCLUDE_URACIL,
        include_repeat:bool = INCLUDE_REPEAT,
        encode_order: list  = ENCODE_ORDER,
        repeat_char: str    = REPEAT_CHAR, # repeated masked regions
        encodex: dict       = FASTA_ENCODEX,
        decodex: dict       = FASTA_DECODEX,
        nts_chars: str      = 'actgu',
        fasta_chars: str    = 'actgrykmswbdhvn-',
        processes: int      = 1
    ):
        '''
        Arguments:
            include_uracil (bool): whether or not uracil should be included in the
                embedding. By default False.

            include_repeat (bool): wehtehr or not repeated masked regions should be
                included in the embedding. By default False.

            encode_order (list): the order in which the channels for the embedding
                should be placed. By default `['a', 'c', 't', 'g', 'u']`.

            repeat_char (str): the character to use for repeated masked regions.
                By default `"."`.

            encodex (dict): the dictionary converting fasta characters to
                nucleotides.

            decodex (dict): the dictionary converting nucleotide characters to
                fasta characters.

            nts_chars (str): valid nucleotide characters: By default: `"actgu"`

            fasta_chars (str): valid fasta characters. By default:
                `"actgrykmswbdhvn-"`

            processes (int): number of CPU cores to use when encoding / decoding.
                By default 1. If set to None, uses all.
        '''
        if processes is None: processes = os.cpu_count()
        self.processes = processes
        self.include_uracil = include_uracil
        self.include_repeat = include_repeat

        if include_uracil and 'u' not in encode_order:
            encode_order += ['u']
        if not include_uracil and 'u' in encode_order:
            encode_order = list(filter(lambda c: not is_uracil(c), encode_order))


        if include_repeat and repeat_char not in encode_order:
            encode_order += [repeat_char]
        self.encode_order = encode_order

        self.repeat_char = repeat_char
        self.encodex = encodex
        self.decodex = decodex
        self.nts_chars = nts_chars
        self.fasta_chars = fasta_chars

    def processes_to_use(self, n):
        '''
        Only intialize at most self.processes, but less if less is needed

        Arguments:
            n (int): number of things to process
        Returns:
            number of processes to use
        '''
        return min(n, self.processes)

    def encode_char(self, char:str):
        '''
        Arguments:
            char (str): a character to encode
        Returns:
            encoding (list): the encoding of the corresponding character.
        '''
        # lowercase letters indicate repeat-masked regions
        is_repeat = char.islower()
        as_lookup = char.lower()

        # what nts are specified by fasta character
        nts = list(self.encodex[as_lookup])

        # filter out uracil
        nts = list(filter(lambda nt: self.include_uracil or not is_uracil(nt), nts))

        encoding = [0 for nt in self.encode_order]
        for nt in nts:
            i = self.encode_order.index(nt)
            encoding[i] = 1


        # indicating if it is a repeat seperately requires custom solution
        if is_repeat and self.include_repeat:
            i = self.encode_order.index(self.repeat_char)
            encoding[i] = 1

        return encoding


    def decode_char(self, char:list):
        '''
        Arguments:
            char (list): an encoding of a character
        Returns:
            fasta (str): the decoded fastas character
        '''
        nts = [
            self.encode_order[i]
            for i, indicator in enumerate(char)
            if indicator > 0
        ]
        is_repeat = self.repeat_char in nts
        clean_nts = [nt for nt in nts if nt != self.repeat_char]
        clean_nts.sort()
        fasta = self.decodex[''.join(clean_nts)]
        if not is_repeat: fasta = fasta.upper()
        return fasta


    def decode(self, emb):
        '''
        Arguments:
            emb (list): a list of embedded characters
        Returns:
            fasta (str): the decoded embedding
        '''
        if self.processes == 1:
            return ''.join(list(map(self.decode_char, emb)))
        else:
            processes = self.processes_to_use(len(emb))
            with Pool(processes=processes) as pool:
                return ''.join(pool.starmap(self.decode_char, [[emb_c] for emb_c in emb]))


    def encode(self, seq):
        '''
        Arguments:
            seq (str): a string of characters to embed
        Returns:
            emb (list): the embedded sequence.
        '''
        if (not is_fasta(seq, self.fasta_chars)):
            raise ValueError('{} is not a valid fasta sequence')
        if self.processes == 1:
            return list(map(self.encode_char, list(seq)))
        else:
            processes = self.processes_to_use(len(list(seq)))
            with Pool(processes=processes) as pool:
                return pool.starmap(self.encode_char, list(seq))



from .defaults import (
    REPEAT_CHAR,
    FASTA_CHARS,
    ENCODE_ORDER,
    INCLUDE_URACIL,
    INCLUDE_REPEAT
)

def is_char(char:str) -> bool:
    '''
    Arguments:
        char (str): a string literal

    Returns:
        result (bool): whether or not the string literal is a single
            character
    '''
    is_str = type(char) is str
    is_sin = len(char) == 1
    return is_str and is_sin

def is_fasta(seq:str, fasta_chars:set=FASTA_CHARS) -> bool:
    '''
    Arguments:
        seq (str): a string literal to test if it represents a
            valid fasta sequence

        fasta_chars (set): the set of valid fasta character

    Returns:
        result (bool): whether or not seq is a valid fasta sequence
    '''
    seq_set = set(''.join(seq.lower()))
    val_set = set(fasta_chars)
    return set(seq_set).issubset(val_set)

def is_uracil(char:str) -> bool:
    '''
    Arguments:
        char (str): a string literal
    Returns:
        result (bool): whether or not char.lower() == 'u'
    '''
    return char.lower() == 'u'

def is_repeat(char):
    '''
    Notes:
        repeated masked regions are indicated by lower case letters
    Arguments:
        char (str): a string literal
    Returns:
        result (bool): whether or not char.islower()
    '''
    # lowercase letters indicate repeat-masked regions
    return char.islower()

def is_char_usable(
    char:str,
    include_uracil: bool  = INCLUDE_URACIL,
    include_repeat: bool  = INCLUDE_REPEAT,
    repeat_char: str      = REPEAT_CHAR,
    fasta_characters: set = FASTA_CHARS
)->bool:
    '''
    Arguments:
        char (str): a string literal

        include_uracil (bool): whether or not uracil should be included in the
            embedding. By default False.

        include_repeat (bool): wehtehr or not repeated masked regions should be
            included in the embedding. By default False.

        repeat_char (str): the character to use for repeated masked regions.
            By default `"."`.

        fasta_chars (set): set of valid fasta characters. By default:
            `"actgrykmswbdhvn-"`
    Returns:
        result (bool):
    '''
    # cond_1 = is_char(char)
    # if not cond_1: return False
    cond_2 = (include_uracil or not is_uracil(char))
    if not cond_2: return False
    cond_3 = (include_repeat or not is_repeat(char, repeat_char))
    return cond_3
    # if not cond_3: return False
    # cond_4 = is_fasta(char, fasta_characters)
    # return cond_4
